package org.mentalog;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.TimeZone;

import org.mentalog.config.BooleanConfigParam;
import org.mentalog.config.ByteArrayConfigParam;
import org.mentalog.config.ConfigParam;
import org.mentalog.config.IntegerConfigParam;
import org.mentalog.config.ListConfigParam;
import org.mentalog.config.LogConfigParam;
import org.mentalog.config.StringConfigParam;
import org.mentalog.config.StringsConfigParam;
import org.mentalog.encoder.Encoder;
import org.mentalog.interceptor.Interceptor;
import org.mentalog.timestamper.Timestamper;
import org.mentalog.util.Benchmarker;
import org.mentalog.util.CheapDateTimeFormatter;
import org.mentalog.util.StringBuilderUtils;
import org.mentalog.util.SystemUtils;


/**
 * An enumeration with a log for each level. The levels are: Debug, Info, Warn, Error, Alert, Fatal.
 * 
 * Each level implement the Logger interface. You can also use the static method createLogger to create other logs not related to any level.
 * 
 * Refer to: http://mentalog.soliveirajr.com for documentations, recipes and more.
 * 
 * @author Sergio Oliveira Jr. - sergio.oliveira.jr@gmail.com
 */
public enum Log implements Logger {

	Debug(36), Info(32), Warn(35), Error(34), Alert(33), Fatal(31);
	
	// /// static stuff //////
	
	static final CheapDateTimeFormatter DATETIME_FORMATTER = new CheapDateTimeFormatter();

	private static final Log DEFAULT_LOG_LEVEL = Log.Info;
	private static final boolean DEFAULT_LOG_CONSOLE = true;
	private static final boolean DEFAULT_LOG_FILE = false;
	private static final Timestamper DEFAULT_TIMESTAMPER = Timestamper.MILLIS;
	private static final TimeZone DEFAULT_TIMEZONE = TimeZone.getDefault();
	private static final String DEFAULT_LOG_DIR = null;
	private static final boolean DEFAULT_SYNCHRONIZED = false;
	private static final boolean DEFAULT_USE_COLORS = false;
	private static final List<Encoder> DEFAULT_ENCODERS = new ArrayList<Encoder>(16);
	private static final boolean DEFAULT_LOG_BYPASSED_CONFIG = false;
	private static final Log DEFAULT_LOG_BYPASSED_CONFIG_LEVEL = Log.Warn;
	private static final boolean DEFAULT_NO_SPACE_AFTER_EQUAL_SIGN = false;
	private static final boolean DEFAULT_NO_SPACE_BETWEEN_OBJECTS= false;
	private static final boolean DEFAULT_LOG_CLASS_AND_LINE_NUMBER = false;
	private static final String DEFAULT_LOG_FILTER = null;
	private static final boolean DEFAULT_LOG_BENCHMARK = false;
	private static final int DEFAULT_LOG_BUFFER_SIZE = 1024 * 10;

	static final byte[] COPYRIGHT = "UNREGISTERED VERSION - Go to http://mentalog.soliveirajr.com/posts/list/0/34.page to register this library.".getBytes();
	
	static {
		
		if (!isRegistered()) {
			System.out.println(new String(COPYRIGHT));
		}
		
		DEFAULT_ENCODERS.add(Encoder.NULL);
		DEFAULT_ENCODERS.add(Encoder.CHAR_SEQUENCE);
		DEFAULT_ENCODERS.add(Encoder.BYTE_OR_CHAR_ARRAY);
		DEFAULT_ENCODERS.add(Encoder.THROWABLE);
	}

	private static BooleanConfigParam logConsole = new BooleanConfigParam("logConsole", DEFAULT_LOG_CONSOLE);
	private static BooleanConfigParam logFile = new BooleanConfigParam("logFile", DEFAULT_LOG_FILE);
	private static LogConfigParam logLevel = new LogConfigParam("logLevel", DEFAULT_LOG_LEVEL);
	private static StringConfigParam logDir = new StringConfigParam("logDir", DEFAULT_LOG_DIR);
	private static BooleanConfigParam isSynchronized = new BooleanConfigParam("logSynchronized", DEFAULT_SYNCHRONIZED);
	private static BooleanConfigParam useColors = new BooleanConfigParam("logColors", DEFAULT_USE_COLORS);
	private static ListConfigParam<Encoder> encoders = new ListConfigParam<Encoder>("logEncoders", "logAddEncoders, logAddEncoder", "logSetEncoders", DEFAULT_ENCODERS);
	private static ConfigParam<Timestamper> timestamper = new ConfigParam<Timestamper>("logTimestamper", DEFAULT_TIMESTAMPER);
	private static ConfigParam<TimeZone> timeZone = new ConfigParam<TimeZone>("logTimeZone", DEFAULT_TIMEZONE);
	private static BooleanConfigParam logBypassedConfig = new BooleanConfigParam("logShowBypassedConfig", DEFAULT_LOG_BYPASSED_CONFIG);
	private static LogConfigParam logBypassedConfigLevel = new LogConfigParam("logShowBypassedConfigLevel", DEFAULT_LOG_BYPASSED_CONFIG_LEVEL);
	private static BooleanConfigParam noSpaceAfterEqualSign = new BooleanConfigParam("logNoSpaceAfterEqualSign", DEFAULT_NO_SPACE_AFTER_EQUAL_SIGN);
	private static BooleanConfigParam noSpaceBetweenObjects = new BooleanConfigParam("logNoSpaceBetweenObjects", DEFAULT_NO_SPACE_BETWEEN_OBJECTS);
	private static BooleanConfigParam logClassAndLineNumber = new BooleanConfigParam("logShowClassAndLineNumber", DEFAULT_LOG_CLASS_AND_LINE_NUMBER);
	private static StringsConfigParam logFilter = new StringsConfigParam("logFilter", DEFAULT_LOG_FILTER);
	private static BooleanConfigParam logBenchmark = new BooleanConfigParam("logBenchmark", DEFAULT_LOG_BENCHMARK);
	private static IntegerConfigParam logBufferSize = new IntegerConfigParam("logBufferSize", DEFAULT_LOG_BUFFER_SIZE);
	
	static {
		
		String configClassName = SystemUtils.getString("mentaLogConfigClass");
		
		if (configClassName != null) {
    		try {
    			Class<?> klass = Class.forName(configClassName);
    			Configuration config = (Configuration) klass.newInstance();
    			ConfigParam.setForceMode(true);
    			config.configure();
    		} catch(Exception e) {
    			throw new IllegalStateException("Cannot load mentalog configuration: " + configClassName);
			} finally {
				ConfigParam.setForceMode(false);
    		}
		}
	}
	
	public static Log fromString(String s) {
		for(Log level : values()) {
			if (level.toString().equalsIgnoreCase(s)) {
				return level;
			}
		}
		return null;
	}
	
	public static int getBufferSize() {
		return logBufferSize.value();
	}
	
	public static void setBufferSize(int size) {
		logBufferSize.set(size);
	}
	
	public static boolean isBenchmark() {
		return logBenchmark.bool();
	}
	
	public static void setBenchmark(boolean flag) {
		logBenchmark.set(flag);
	}
	
	public static String[] getFilter() {
		return logFilter.values();
	}
	
	public static void setFilter(String filter) {
		logFilter.set(filter);
	}
	
	public static boolean isShowClassAndLineNumber() {
		return logClassAndLineNumber.bool();
	}
	
	public static void setShowClassAndLineNumber(boolean flag) {
		logClassAndLineNumber.set(flag);
	}

	public static boolean isNoSpaceBetweenObjects() {
		return noSpaceBetweenObjects.bool();
	}

	public static void setNoSpaceBetweenObjects(boolean flag) {
		noSpaceBetweenObjects.set(flag);
	}

	public static boolean isNoSpaceAfterEqualSign() {
		return noSpaceAfterEqualSign.bool();
	}

	public static void setNoSpaceAfterEqualSign(boolean flag) {
		noSpaceAfterEqualSign.set(flag);
	}
	
	public static List<Encoder> getEncoders() {
		return encoders.values();
	}
	
	public static void add(Encoder encoder) {
		encoders.add(encoder);
	}
	
	public static void setEncoders(List<Encoder> encs) {
		encoders.set(encs);
	}
	
	public static boolean isColors() {
		return useColors.bool();
	}
	
	public static void setColors(boolean b) {

		boolean oldValue = useColors.bool();

		useColors.set(b);

		if (b != oldValue) {
			closeAll();
		}
	}
	
	public static boolean isSynchronized() {
		return isSynchronized.bool();
	}
	
	public static void setSynchronized(boolean flag) {

		boolean oldValue = isSynchronized.bool();

		isSynchronized.set(flag);

		if (flag != oldValue) {
			closeAll();
		}
	}

	public static boolean isConsole() {
		return logConsole.bool();
	}

	public static void setLevel(Log level) {
		logLevel.set(level);
	}

	public static Log getLevel() {
		return logLevel.value();
	}
	
	public static void setShowBypassedConfig(boolean flag, Log level) {
		logBypassedConfig.set(flag);
		logBypassedConfigLevel.set(level);
	}
	
	public static boolean isShowBypassedConfig() {
		return logBypassedConfig.bool();
	}
	
	public static Log getShowBypassedConfigLevel() {
		return logBypassedConfigLevel.value();
	}

	public static Timestamper getTimestamper() {
		return timestamper.value();
	}

	public static void setConsole(boolean flag) {

		if (logFile.isForced()) {
				ConfigParam.checkAndLog("logConsole", flag, logConsole.bool());
				return;
		}

		boolean oldValue = logConsole.bool();

		logConsole.set(flag);
		logFile.set(!flag);

		if (logConsole.bool() != oldValue) {
			closeAll();
		}
	}
	
	public static void setFile(boolean flag) {

		if (logConsole.isForced()) {
			ConfigParam.checkAndLog("logFile", flag, logFile.bool());
			return;
		}

		boolean oldValue = logFile.bool();

		logFile.set(flag);
		logConsole.set(!flag);

		if (logFile.bool() != oldValue) {
			closeAll();
		}
	}

	public static boolean isFile() {
		return logFile.bool();
	}

	public static void setTimestamper(Timestamper ts) {
		if (ts == null) {
			throw new LogException("Cannot set timestamper to null!");
		}
		
		Timestamper oldValue = timestamper.value();
		
		timestamper.set(ts);
		
		if (timestamper.value() != oldValue) {
			closeAll();
		}
	}
	
	public static void setTimeZone(TimeZone tz) {
		if (tz == null) {
			throw new LogException("Cannot set timezone to null!");
		}
		
		TimeZone oldValue = timeZone.value();
		
		timeZone.set(tz);

		if (timeZone.value() != oldValue) {
			DATETIME_FORMATTER.changeTimeZone(tz);
		}
	}

	public static TimeZone getTimeZone() {
		return timeZone.value();
	}

	public static void setDir(String dir) {

		String oldValue = logDir.str();

		logDir.set(dir);

		if (!logDir.str().equals(oldValue)) {
			closeAll();
		}
	}

	public static String getDir() {
		return logDir.str();
	}

	///////////////////////////

	private final String prefix;
	private final ByteArrayConfigParam color;
	private final StringConfigParam logFilename;
	private final BooleanConfigParam enabled;
	
	private Logger logger = null;

	private Log(int color) {
		String name = this.toString();
		prefix = name.toUpperCase();

		String defaultFilename = name.toUpperCase() + ".log";
		String logFileArgumentName = "log" + name + "Filename";
		this.logFilename = new StringConfigParam(logFileArgumentName, defaultFilename);
		
		String colorArgumentName = name.toLowerCase() + "Color";
		this.color = new ByteArrayConfigParam(colorArgumentName, getBytesFromColor(color));
		
		String enabledArgumentName = name.toLowerCase() + "Enabled";
		this.enabled = new BooleanConfigParam(enabledArgumentName, null);
	}

	private static byte[] getBytesFromColor(int color) {
		return ("\033[1;" + String.valueOf(color) + "m").getBytes();
	}
	
	@Override
	public void addInterceptor(Interceptor interceptor) { 
		Logger logger = getLogger();
		if (logger != null) {
			logger.addInterceptor(interceptor);
		}
	}
	
	@Override
	public void removeInterceptor(Interceptor interceptor) {
		Logger logger = getLogger();
		if (logger != null) {
			logger.removeInterceptor(interceptor);
		}
	}
	
	public void setColor(int color) {
		if (color < 30 || color > 37) {
			throw new LogException("Color can only be in the range 30-37 !");
		}
		
		byte[] oldValue = this.color.bytes();
		
		this.color.set(getBytesFromColor(color));
		
		if ((isConsole() || !isFile()) && !Arrays.equals(oldValue, this.color.bytes())) {
			closeAll();
		}
	}
	
	public void setFilename(String filename) {
		
		String oldValue = logFilename.str();

		logFilename.set(filename);

		if (!oldValue.equals(logFilename.str())) {
			closeAll();
		}
	}

	public static void closeAll() {
		for(Log level: values()) {
			level.close();
		}
	}
	
	public static void add(Interceptor interceptor) {
		for(Log log: values()) {
			log.addInterceptor(interceptor);
		}
	}
	
	public static void remove(Interceptor interceptor) {
		for(Log log : values()) {
			log.removeInterceptor(interceptor);
		}
	}
	
	@Override
	public final void close() {
		if (logger != null) {
			logger.close();
			logger = null;
		}
	}
	
	public void logBenchmark() {
		StringBuilder sb = new StringBuilder(128);
		sb.append("Benchmarker=").append(Log.isBenchmark() ? "ON" : "OFF");
		sb.append(" Results=[").append(Benchmarker.instance().results()).append("]");
		log(sb);
	}
	
	public boolean isOn(Log level) {
		return this.ordinal() >= level.ordinal();
	}
	
	final Logger getLogger() {
		if (isSynchronized.bool()) {
			synchronized(this) {
				return getLoggerImpl();
			}
		} else {
			return getLoggerImpl();
		}
	}

	private final Logger getLoggerImpl() {
		if (isEnabled()) {
			if (logger == null) {
				logger = createLogger();
			}
			return logger;
		} else {
			close();
			return null;
		}
	}
	
	public static void enableAll(Boolean flag) {
		for (Log log : values()) {
			log.enable(flag);
		}
	}

	public final void enable(Boolean flag) {
		enabled.set(flag);
	}

	@Override
	public final boolean isEnabled() {

		if (enabled.value() != null && enabled.bool()) {
			return true;
		}
		
		if (enabled.value() != null && !enabled.bool()) {
			return false;
		}
		
		if (Log.getLevel() == null) {
			return false;
		}

		if (this.isOn(Log.getLevel())) {
			return true;
		}
		
		return false;
	}

	public static Logger createLogger(String dir, String filename, Timestamper timestamper, boolean isSynchronized) {
		return createLogger(dir, filename, timestamper, isSynchronized, encoders.values());
	}
	
	public static Logger createLogger(String dir, String filename, boolean isSynchronized) {
		return createLogger(dir, filename, DEFAULT_TIMESTAMPER, isSynchronized, encoders.values());
	}
	
	private static List<Encoder> convertToArrayList(List<Encoder> encoders) {
		List<Encoder> list = new ArrayList<Encoder>(encoders.size());
		list.addAll(encoders);
		return list;
	}

	public static Logger createLogger(String dir, String filename, boolean isSynchronized, List<Encoder> encoders) {
		return createLogger(dir, filename, DEFAULT_TIMESTAMPER, isSynchronized, encoders);
	}

	public static Logger createLogger(String dir, String filename, Timestamper timestamper, boolean isSynchronized, List<Encoder> encoders) {

		if (!(encoders instanceof ArrayList)) {
			// we will iterate using a for loop, not to waste iterators...
			// so make sure we have an ArrayList here not to be slow
			encoders = convertToArrayList(encoders);
		}

		try {
			createDir(dir);
			return new DefaultLogger(dir, filename, timestamper, isSynchronized, encoders, null);
		} catch(Exception e) {
			throw new LogException("Cannot create logger!", e);
		}
	}
	
	private static void createDir(String dir) {
		if (dir != null && !dir.equals("")) {
			File f = new File(dir);
			if (!f.exists()) {
				f.mkdirs();
			}
		}
	}
	
	private Logger createLogger() {
		
		try {

			if (logConsole.isForcedTrue() && logFile.isForcedTrue()) {
				throw new LogException("Cannot configure logConsole and logFile at the same time!");
			}
			
			if (logConsole.isForcedTrue() || logFile.isForcedFalse() || logConsole.bool() || !logFile.bool()) {
				
				byte[] c = null;
				
				if (useColors.bool()) {
					
					c = this.color.bytes();
				}

				return new DefaultLogger(System.out, prefix, timestamper.value(), isSynchronized.bool(), c, encoders.values(), this);

			} else {

				String fn = logFilename.str();

				String dir = logDir.str();

				createDir(dir);
				
				return new DefaultLogger(dir, fn, timestamper.value(), isSynchronized.bool(), encoders.values(), this);
			}

		} catch (IOException e) {
			throw new LogException("Error creating log!", e);
		}
	}
	
	@Override
	public final void roll() {

		Logger logger = getLogger();
		
		if (logger != null) {
			logger.roll();
		}
	}

	@Override
	public final void log(Object... objects) {
		
		Logger logger = getLogger();

		if (logger != null) {
			logger.log(objects);
		}
	}
	
	public static void main(String[] args) {
		
		// command-line: java -cp target/classes/ -DuseColors=true org.mentalog.Log

		String s = "Hello MentaLog!";
		
		Log.setLevel(Log.Debug);

		for(Log log : Log.values()) {
			log.log(s);
		}
	}
	
	static boolean isRegistered() {
		
		try {
			Class.forName("org.mentalog.License");
			return true;
		} catch(Exception e) {
			return false;
		}
	}
	
	/////// VARARGS /////
	
	@Override
	public final void log(Object s1) {
		Logger logger = getLogger();

		if (logger != null) {
			logger.log(s1);
		}
	}
	
	@Override
	public final void log(Object s1, Object s2) {
		Logger logger = getLogger();

		if (logger != null) {
			logger.log(s1, s2);
		}
	}
	
	@Override
	public final void log(Object s1, Object s2, Object s3) {
		Logger logger = getLogger();

		if (logger != null) {
			logger.log(s1, s2, s3);
		}
	}

	@Override
	public final void log(Object s1, Object s2, Object s3, Object s4) {
		Logger logger = getLogger();

		if (logger != null) {
			logger.log(s1, s2, s3, s4);
		}
	}

	@Override
	public final void log(Object s1, Object s2, Object s3, Object s4, Object s5) {
		Logger logger = getLogger();

		if (logger != null) {
			logger.log(s1, s2, s3, s4, s5);
		}
	}

	@Override
	public final void log(Object s1, Object s2, Object s3, Object s4, Object s5, Object s6) {
		Logger logger = getLogger();

		if (logger != null) {
			logger.log(s1, s2, s3, s4, s5, s6);
		}
	}

	@Override
	public final void log(Object s1, Object s2, Object s3, Object s4, Object s5, Object s6, Object s7) {
		Logger logger = getLogger();

		if (logger != null) {
			logger.log(s1, s2, s3, s4, s5, s6, s7);
		}
	}

	@Override
	public final void log(Object s1, Object s2, Object s3, Object s4, Object s5, Object s6, Object s7, Object s8) {
		Logger logger = getLogger();

		if (logger != null) {
			logger.log(s1, s2, s3, s4, s5, s6, s7, s8);
		}
	}

	@Override
	public final void log(Object s1, Object s2, Object s3, Object s4, Object s5, Object s6, Object s7, Object s8, Object s9) {
		Logger logger = getLogger();

		if (logger != null) {
			logger.log(s1, s2, s3, s4, s5, s6, s7, s8, s9);
		}
	}

	@Override
	public final void log(Object s1, Object s2, Object s3, Object s4, Object s5, Object s6, Object s7, Object s8, Object s9, Object s10) {
		Logger logger = getLogger();

		if (logger != null) {
			logger.log(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10);
		}
	}

	@Override
	public final void log(Object s1, Object s2, Object s3, Object s4, Object s5, Object s6, Object s7, Object s8, Object s9, Object s10, Object s11) {
		Logger logger = getLogger();

		if (logger != null) {
			logger.log(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11);
		}
	}

	@Override
	public final void log(Object s1, Object s2, Object s3, Object s4, Object s5, Object s6, Object s7, Object s8, Object s9, Object s10, Object s11, Object s12) {
		Logger logger = getLogger();

		if (logger != null) {
			logger.log(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12);
		}
	}

	@Override
	public final void log(Object s1, Object s2, Object s3, Object s4, Object s5, Object s6, Object s7, Object s8, Object s9, Object s10, Object s11, Object s12, Object s13) {
		Logger logger = getLogger();

		if (logger != null) {
			logger.log(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13);
		}
	}

	@Override
	public final void log(Object s1, Object s2, Object s3, Object s4, Object s5, Object s6, Object s7, Object s8, Object s9, Object s10, Object s11, Object s12, Object s13, Object s14) {
		Logger logger = getLogger();

		if (logger != null) {
			logger.log(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14);
		}
	}

	@Override
	public final void log(Object s1, Object s2, Object s3, Object s4, Object s5, Object s6, Object s7, Object s8, Object s9, Object s10, Object s11, Object s12, Object s13, Object s14, Object s15) {
		Logger logger = getLogger();

		if (logger != null) {
			logger.log(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15);
		}
	}

	@Override
	public final void log(Object s1, Object s2, Object s3, Object s4, Object s5, Object s6, Object s7, Object s8, Object s9, Object s10, Object s11, Object s12, Object s13, Object s14, Object s15, Object s16) {
		Logger logger = getLogger();

		if (logger != null) {
			logger.log(s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13, s14, s15, s16);
		}
	}

	// Stuff for the StringBuilderUtils:

	public final static StringBuilder to_sb(int x) {
		return StringBuilderUtils.to_sb(x);
	}

	public final static StringBuilder to_sb(long x) {
		return StringBuilderUtils.to_sb(x);
	}

	public final static StringBuilder to_sb(byte b) {
		return StringBuilderUtils.to_sb(b);
	}

	public final static StringBuilder to_sb(short s) {
		return StringBuilderUtils.to_sb(s);
	}

	public final static StringBuilder to_sb(double x) {
		return StringBuilderUtils.to_sb(x);
	}

	public final static StringBuilder to_sb(boolean b) {
		return StringBuilderUtils.to_sb(b);
	}

	public final static StringBuilder to_sb(float f) {
		return StringBuilderUtils.to_sb(f);
	}

	public final static StringBuilder to_sb(char c) {
		return StringBuilderUtils.to_sb(c);
	}

}